# -*- coding: utf-8 -*-

"""
kaoru.commands
~~~~~~~~

The very basics

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import Update
from .. import db
from .. import log
from .. import security

def bot_command(command_func):
    """
    A decorator whose purpose is to stack middleware
    routines on top of its target function
    """

    # the actual wrapper function
    def _wrapper(bot, update):
        ###################################
        # all middleware can be run in here
        ###################################
        if isinstance(update, Update):
            username = update.message.from_user.username
            userid = update.message.from_user.id
            command = update.message.text
            log.msg_debug("[{}:{}] Received command: '{}'".format(
                    username, userid, command
                )
            )

            #########################
            # perform security checks
            #########################
            try:
                security.check_update(update)
            except security.SecurityException as sec_except:
                log.msg_err(sec_except)
                return

            ###################################################
            # Grab the update and append it to database to keep
            # constant track of what this bot is getting
            ###################################################
            db.insert_update(update)
        else:
            # I have received a string command
            log.msg_debug("Received command: '{}'".format(update))

        # Finally, execute the intended command
        command_func(bot, update)

    # give the thing back!
    return _wrapper



# -*- coding: utf-8 -*-

"""
kaoru.procutils
~~~~~~~~

Utilities for executing external processes

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import envoy

from . import log
from . import config


def _proc_exec(exec_func):
    """Run a process, asynchronously"""

    def _exec_func(cmd):
        log.msg_debug("exec: {}".format(cmd))
        if not config.get('dry_run'):
            return exec_func(cmd)
        return None
    return _exec_func

@_proc_exec
def proc_exec_async(cmd):
    """Run a process, asynchronously"""

    envoy.connect(cmd)
    return None

@_proc_exec
def proc_exec(cmd):
    """Run a process, synchronously"""

    return envoy.run(cmd)

def proc_select(exec_list):
    """Select the first executable found from a list

    Select the preferred executable for a command, if user_exec is None,
    this will check for every executable in exec_list in PATH.
    If none is found, it will return None

    :param exec_list: list of executable nominees
    """

    # if the user didn't select any particular
    # executable for a command, it'll be selected from
    # exec_list
    for prog in exec_list:
        r = envoy.run("which {}".format(prog))
        if r.status_code == 0:
            return r.std_out[:-1]
    return None



# -*- coding: utf-8 -*-

"""
kaoru.commands.start
~~~~~~~~

/start command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import Update, ParseMode, Emoji
from . import bot_command
from .. import utils

# /start command
@bot_command
def _cmd_handler(bot, update):
    # set the user name
    username = 'user'
    if isinstance(update, Update):
        username = update.message.from_user.first_name

    # the actual message to semt
    start_msg = [
        '_Hello world!_  ... Oh!, I meant: Hello {} ... {}'
        .format(username, Emoji.FLUSHED_FACE),
        'I can\'t promise I won\'t break, '
        'but I\'ll put my best to assist you!',
        'Anyways, *let\'s do this!, shall we?*, start by asking me for /help'
    ]

    # and send the thing
    for msg in start_msg:
        utils.echo_msg(
            bot, update, msg, parse_mode=ParseMode.MARKDOWN
        )

cmd_list = None  #
desc = 'Start me up'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'start'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.hello
~~~~~~~~

/hello command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import Update
from .. import utils
from . import bot_command

_greetings = [
    # English
    'Affirmative Dave, I read you',
    'Hello world!',

    # Spanish
    'Hola',

    # Arabic
    'أهلاً و سهلاً',

    # Mandarin
    '你好',

    # Corsican
    'Salute',

    # French
    'Salut', 'Bonjour!, est-ce que vous allez bien?',

    # Danish
    'Hej',

    # German
    'Hallo',
    'Guten tag!',

    # Italian
    'Ciao',

    # Japanese
    '今日は',

    # Klingon
    'nuqneH',

    # Farsi
    'سلام',

    # Turkish
    'Merhaba',
]

# /hello command
@bot_command
def _cmd_handler(bot, update):
    """a rather simple ping command"""

    if isinstance(update, Update):
        utils.echo_msg( bot, update, utils.select_rand_str(_greetings))
    else:
        utils.echo_msg(bot, update, utils.select_rand_str(_greetings))

desc = 'See if I "live"'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'hello'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru
~~~~~~~~

The very basics

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

PKG_URL = 'https://github.com/axltxl/kaoru'
__name__ = 'kaoru'
__author__ = 'Alejandro Ricoveri'
__version__ = '0.5.0'
__licence__ = 'MIT'
__copyright__ = 'Copyright (c) Alejandro Ricoveri'



# -*- coding: utf-8 -*-

"""
kaoru.commands.screenlock
~~~~~~~~

/screenlock implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import Update
from .. import utils
from .. import config
from .. import log
from . import bot_command
from ..procutils import proc_exec_async

# /screenlock command:
@bot_command
def _cmd_handler(bot, update):
    """
    it basically runs a screen locker
    """

    # check for executables set for commands
    screenlock_exec = config.get('screenlock_cmd')
    if screenlock_exec is None:
        screenlock_exec = proc_select([
            'xlock', 'xscreensaver', 'i3lock'
            ],
        )
    # only execute the thing if existent
    if screenlock_exec is None:
        err_msg = "A suitable 'screenlock_cmd' has bot been found."
        utils.echo_msg(err_msg)
        log.msg_err(err_msg)
        return

    proc_exec_async(screenlock_exec)
    utils.echo_msg(bot, update, "Your screen(s) are now LOCKED")

desc = 'See if I "live"'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'screenlock'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands
~~~~~~~~

Command handlers

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from . import log
from .commands import (
    about,
    start,
    hello,
    help,
    screenlock,
    screenshot,
    dryrun,
    cancel,
    poweroff,
    suspend,
    hibernate,
    reboot,
    unknown,
)

# list of available commands
_commands = [
    (about.cmd_str, about.desc, about.cmd_handler),
    (start.cmd_str, start.desc, start.cmd_handler),
    (hello.cmd_str, hello.desc, hello.cmd_handler),
    (help.cmd_str, help.desc, help.cmd_handler),
    (screenlock.cmd_str, screenlock.desc, screenlock.cmd_handler),
    (screenshot.cmd_str, screenshot.desc, screenshot.cmd_handler),
    (dryrun.cmd_str, dryrun.desc, dryrun.cmd_handler),
    (cancel.cmd_str, cancel.desc, cancel.cmd_handler),
    (reboot.cmd_str, reboot.desc, reboot.cmd_handler),
    (poweroff.cmd_str, poweroff.desc, poweroff.cmd_handler),
    (suspend.cmd_str, suspend.desc, suspend.cmd_handler),
    (hibernate.cmd_str, hibernate.desc, hibernate.cmd_handler),
]

# set cmd list to be used by /help command
help.cmd_list = _commands

def _print_cmd_desc(commands):
    """Print command descriptions"""
    log.msg_warn("You will need to register my commands with my @BotFather")
    log.msg_warn("Ask him to /setcommands and after you")
    log.msg_warn("have mentioned me, you can paste the following:")
    log.msg("")
    for command, desc, handler in commands:
        print("{} - {}".format(command, desc))
    log.msg("")

def get_list():
    """List of available commands"""

    return _commands

def register_commands(updater, dispatcher):
    """register each and every command this bot is going to process"""

    # register every command there is
    for command, desc, handler in _commands:
        dispatcher.addTelegramCommandHandler(command, handler)
        dispatcher.addStringCommandHandler(command, handler)
        log.msg_debug("/{}: command registered".format(command))

    # and the default one as well ...
    log.msg_debug('Registering default handlers')
    dispatcher.addUnknownTelegramCommandHandler(unknown.cmd_handler)

    # print command descriptions
    _print_cmd_desc(_commands)




# -*- coding: utf-8 -*-

"""
kaoru.config
~~~~~~~~

Configuration goes in here

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import yaml
from schema import Schema, Optional, And

from . import log

# Default set of options
_options = {
    'token': None,
    'queue_reboot': False,
    'queue_poweroff': False,
    'discard_on_startup': True,
    'reboot_delay': 0,
    'poweroff_delay': 0,
    'screenlock_cmd': None,
    'dry_run': False,
    'strict': False,
    'masters': [],
    'show_hostname': False,
    'cli': False,
}
_schema = None


def init(*, config_file):
    """Configuration file

    Its role's very simple, to load
    a YAML file, validate it and from there
    set options to be used by this program

    :param config_file: path to configuration file
    """

    global _schema
    global _options

    # Validation schema for each option
    _schema = Schema({
        Optional('token'): str,
        Optional('strict'): bool,
        Optional('discard_on_startup'): bool,
        Optional('masters'): lambda x: isinstance(x,list) and all([isinstance(y,str) for y in x]),
        Optional('dry_run'): bool,
        Optional('show_hostname'): bool,
        Optional('reboot_delay'): int,
        Optional('screenlock_cmd'): str,
        Optional('poweroff_delay'): int,
        })

    # Load YAML file and validate it
    try:
        with open(config_file, "r") as file:
            data = _schema.validate(yaml.load(file))
            for key, value in data.items():
                # overwrite defaults with values taken
                # from the configuration file
                _list_merge(data, _options)
    except FileNotFoundError as fnfe:
        log.msg_warn(
            '{}: configuration file not found, '
            'proceeding with default values'.format(config_file)
        )


def _list_merge(src, dest):
    """
    Merge the contents coming from src into dest

    :param src: source dictionary
    :param dest: destination dictionary
    """
    for k in src:
        if type(src[k]) != dict:
            dest[k] = src[k]
        else:
            # ---
            # src could have a key whose value is a list
            # and does not yet exist on dest
            if k not in dest:
                dest[k] = {}
            _list_merge(src[k], dest[k])

def options():
    """Get the whole options set"""
    return _options

def get(key):
    if key in _options:
        return _options[key]
    return None

def set(key, value):
    global _options
    if key in _options:
        _options[key] = value




# -*- coding: utf-8 -*-

"""
kaoru.app
~~~~~~~~

Main module

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import sys
import traceback
import os
import signal
import time
from telegram import Updater
from telegram import Bot
from docopt import docopt

from . import __version__ as version
from . import PKG_URL as pkg_url
from . import __name__ as pkg_name
from . import log
from . import command
from . import config
from . import security
from . import cli
from . import utils
from . import db

# telegram objects to be used
_tg_dispatcher = None
_tg_updater = None

##################################################
# Configuration file defaults
# default config directory should be XDG-compliant
##################################################
_config_dir_default = None
_config_file_default = None


##################################################
# Database file defaults
##################################################
_db_dir = None
_db_file = None

def _config_init(args):
    """Initialise the configuration file

        :param args: command line arguments
    """

    config_file = args['--config']
    if config_file is None:
        config_file = "{}/{}".format(_config_dir_default, _config_file_default)
    log.msg("Reading configuration file at: {}".format(config_file))
    config.init(config_file=config_file)

    #########################################################
    # Telegram Bot API token:
    # It can be obtained by either via the configuration file
    # or the environment variable TG_TOKEN
    #########################################################
    if config.get('token') is None:
        token = os.getenv('TG_TOKEN')
        if token is None:
            raise SystemError('A Telegram Bot API token is mandatory for this thing to work!')
        else:
            log.msg_debug('got Telegram Bot API token from environment variable')
        config.set('token', token)
    else:
        log.msg_debug('got Telegram Bot API token from configuration file')

    # whether to set dry run mode
    config.set('dry_run', args['--dry-run'])

    # activate cli mode
    config.set('cli', args['--interactive'])

    # print final configuration
    log.msg_debug("Configuration settings are the following:")
    log.msg_debug("-----------------------------------------")
    cfg_opts = config.options().copy() # since token is "mangled before showing
    cfg_opts['token'] = _mangle_token(cfg_opts['token'])
    for key, value in cfg_opts.items():
        log.msg_debug("{:16} => {}".format(key, value))
    log.msg_debug("-----------------------------------------")

def _mangle_token(token):
    """Slaughter a token for public exposure"""
    _tl = len(token)//2
    return "{}{}".format('*' * (_tl), token[_tl+1:])

def _log_init(log_file, log_lvl):
    """Initialise log file"""

    if log_file is None:
        log_file = "{}/{}".format(_config_dir_default, log.LOG_FILE_DEFAULT)
    log.init(log_file=log_file, threshold_lvl=int(log_lvl))

def _db_init():
    """Intitialise database file"""
    db.init(db=_db_file)

def _base_dirs_init():
    """Initialise base configuration directory

    This base directory is used to read configuration (if --config is not set)
    and also used for appending a log file (if --log-file is not set)
    """

    global _config_dir_default, _config_file_default

    # This configuration directory should be XDG-compliant, but not expected
    config_dir = os.getenv('XDG_CONFIG_HOME') # default base directory
    if config_dir is None:
        config_dir = "{}/.config".format(os.getenv('HOME'))

    # set base configuration directory and corresponding file as well
    _config_dir_default = "{}/{}".format(config_dir, pkg_name)
    _config_file_default = "{}.conf".format(pkg_name)

    # Create configuration directory if it does not exist
    if not os.path.exists(_config_dir_default):
        log.msg_warn(
            "Configuration directory '{}' does not exist, creating it ..."
            .format(_config_dir_default)
        )
        # create the actual default configuration directory
        os.makedirs(_config_dir_default)

    # Define database directory
    global _db_dir, _db_file
    data_dir = os.getenv('XDG_DATA_HOME')
    if data_dir is None:
        data_dir = "{}/.local/share".format(os.getenv('HOME'))
    _db_dir = "{}/{}".format(data_dir, pkg_name)
    log.msg_debug("database directory is located at '{}'".format(_db_dir))

    # define database file
    _db_file = "{}/{}.db".format(_db_dir, pkg_name)

    # Check for database directory
    if not os.path.exists(_db_dir):
        log.msg_warn(
            "Database directory '{}' does not exist, creating it ..."
            .format(_db_dir)
        )
        # create the actual default configuration directory
        os.makedirs(_db_dir)


def init(argv):
    """Usage: kaoru [options]

    -L LVL --log-level LVL  Verbosity level on output [default: 0]
    -i --interactive        CLI mode
    -l --log-file FILE      Log file
    -c FILE --config FILE   Configuration file to use
    -d --dry-run            Dry run mode (don't do anything)
    """

    args = docopt(init.__doc__, argv=argv[1:], version=version)

    # This baby will handle UNIX signals
    signal.signal(signal.SIGINT, _handle_signal)
    signal.signal(signal.SIGTERM, _handle_signal)


    # initialise randomizer
    utils.random_seed()

    # Intialialise base directories, first of all
    _base_dirs_init()

    # initialize log
    _log_init(args['--log-file'], args['--log-level'])

    # show splash
    _splash()

    # initialise configuration file
    _config_init(args)

    # initialise database
    _db_init()

    # check for strict mode
    if config.get('strict'):
        log.msg_warn("Strict mode has been enforced")
        security.check_masters(config.get('masters'))

    # give back the list of arguments captured by docopt
    return args

def handle_except(e):
    """
    Handle (log) any exception

    :param e: exception to be handled
    """
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    log.msg_err("Unhandled {e} at {file}:{line}: '{msg}'" .format(
        e=exc_type.__name__, file=fname,
        line=exc_tb.tb_lineno, msg=e))
    log.msg_err(traceback.format_exc())
    log.msg_err("An error has occurred!. "
                "For more details, review the logs.")
    return 1

def _handle_error(bot, update, error):
    raise error

def _get_last_update_id():
    """Get the latest update id"""

    # first of all, get latest update id on record
    # if there's none, just set it to zero
    last_update_id = db.get_last_update_id()
    if last_update_id is None:
        return 0

    #######################
    # initial update offset
    #######################
    last_update_id += 1

    ############################################################
    # Discarding updates on startup:
    # ------------------------------
    # Namely, to take all updates since last_update_id
    # and "discard" them, which means that, once updates
    # are captured, another call to getUpdates is done with
    # an offset higher than the latest update id, resulting
    # in all previous updates to be confirmed on backend
    # This is done to make sure that this bot gets updates older
    # than its latest update id on record
    # In this way, if the user sends commands before this bot
    # starts to run, those updates won't be taken into account
    ############################################################
    if config.get('discard_on_startup'):
        log.msg_warn(
            'Any previously unconfirmed updates (newer than {}) '.format(last_update_id)
            + 'will be discarded on startup'
        )
        log.msg_warn(
            'Proceeding to discard pending updates newer than {}'
            .format(last_update_id)
        )

        # create a temporary bot for the dirty job
        bot = Bot(token=config.get('token'))
        log.msg_debug(
            "telegram.Bot.getUpdates(offset={})" .format(last_update_id)
        )

        # get updates equal or higher than last_update
        updates = bot.getUpdates(offset=last_update_id)
        if len(updates):
            log.msg_warn(
                '{} pending update(s) captured and discarded ...'
                .format(len(updates))
            )

            # get the latest update if
            last_update_id = updates[-1].update_id + 1

            # and force a getUpdates call to Bot API
            # to declare all previous updates as confirmed
            bot.getUpdates(offset=last_update_id)
        else:
            log.msg_warn(
                "No unconfirmed update(s) found newer than {}"
                .format(last_update_id)
            )

    # and return the actual latest update id
    return last_update_id


##################
# The main "thing"
##################
def start():
    """Start the application itself """

    # ... and as well a few other things that are necessary
    global _tg_updater, _tg_dispatcher

    #########################
    # create Telegram updater
    #########################
    _tg_updater = Updater(token=config.get('token'))

    # ... and get its dispatcher, which it's manipulated
    # on further code
    _tg_dispatcher = _tg_updater.dispatcher

    ###############
    # error handler
    # TODO: causing errors
    ###############
    # log.msg_debug('Registering error handler')
    # _tg_dispatcher.addErrorHandler(_handle_error)

    # Register callbacks for each supported command in this bot
    command.register_commands(_tg_updater, _tg_dispatcher)

    #############################################################
    # set initial last update id:
    # By setting it, the updater's bot will start getting updates
    # with an offset of last_update_id instead of default 0,
    # consequently, only unconfirmed updates greater or equal
    # than last_update_id will be taken into account:
    # see: https://github.com/python-telegram-bot/python-telegram-bot/blob/master/telegram/updater.py#L171
    # see: http://python-telegram-bot.readthedocs.org/en/latest/telegram.bot.html#telegram.bot.Bot.getUpdates
    #############################################################
    _tg_updater.bot.last_update_id = _get_last_update_id()
    log.msg_debug(
        "update offset => {}" .format(_tg_updater.bot.last_update_id)
    )

    # reset logger interface so we can track threads
    # create by our telegram client
    _tg_updater.logger = log.get_logger()

    ####################################
    # Start listening for actual updates
    ####################################
    update_queue = _tg_updater.start_polling(
        poll_interval=0.5,
        timeout=5,
        network_delay=10
    )

    # Mark the beginning of everything
    log.msg("Waiting for updates ...")

    # Start CLI-Loop
    if config.get('cli'):
        cli.prompt_loop(_tg_dispatcher, update_queue)
    else:
        while True:
            time.sleep(1)


def _handle_signal(signum, frame):
    """
    UNIX signal handler
    """
    # Raise a SystemExit exception
    sys.exit(1)

def shutdown():
    """
    Cleanup
    """

    # Make sure the updater has stopped
    if _tg_updater is not None:
        log.msg_debug('Stopping updater ..')
        _tg_updater.stop()

    log.msg("Exiting ...")

def _splash():
    """Print the splash"""
    splash_title = "{pkg} [{version}] - {url}".format(
        pkg=pkg_name, version=version, url=pkg_url)
    log.to_stdout(splash_title, colorf=log.yellow, bold=True)
    log.to_stdout('-' * len(splash_title), colorf=log.yellow, bold=True)
    log.to_stdout(
        "Please, report issues to {}/issues"
        .format(pkg_url), colorf=log.yellow, bold=True
    )




# -*- coding: utf-8 -*-

"""
kaoru.utils
~~~~~~~~

Utilities

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import socket
import random
import time

from telegram import Update, ChatAction
from . import log
from . import config

def random_seed():
    """Seed the randomizer"""
    random.seed(int(time.time()))

def echo_msg(bot, update, msg, **kwargs):
    if isinstance(update, Update):
        # some basic info about the user
        userid = update.message.from_user.id
        username = update.message.from_user.username
        chat_id = update.message.chat_id

        # log the thing
        log.msg("echo message [{}@{}]: {}".format(username, userid, msg))

        # reformat message so it includes this machine's host name
        show_hostname = config.get('show_hostname')
        dry_run = config.get('dry_run')
        parse_mode = kwargs.get('parse_mode')
        if (show_hostname or dry_run) and parse_mode is None:
            tag = ""
            if show_hostname:
                tag = socket.gethostname()
            if dry_run:
                if len(tag):
                    tag += ":dr"
                else:
                    tag = "dr"
            tag = "[{}]".format(tag)
            msg = "{} {}".format(tag, msg)

        # send the actual message
        bot.sendChatAction(chat_id=chat_id, action=ChatAction.TYPING)
        bot.sendMessage(
            chat_id=chat_id,
            text=msg,
            **kwargs
        )
    else:
        log.msg("{} > {}".format(bot.username, msg))

def select_rand_str(candidates):
    """select a random string from a list

    :candidates: list of string candidates
    :returns: str

    """
    return candidates[random.randrange(0, len(candidates))]



# -*- coding: utf-8 -*-

"""
kaoru.__main__
~~~~~~~~

Main entry point

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import sys

from docopt import DocoptExit

from kaoru.app import handle_except, init, shutdown, start

def main(argv=None):
    """
    This is the main thread of execution

    :param argv: list of command line arguments
    """
    # Exit code
    exit_code = 0

    # First, we change main() to take an optional 'argv'
    # argument, which allows us to call it from the interactive
    # Python prompt
    if argv is None:
        argv = sys.argv

    try:
        # Bootstrap
        options = init(argv)

        # start the thing!
        start()

    except DocoptExit as dexcept:
        # Deal with wrong arguments
        print(dexcept)
        exit_code = 1
    except Exception as e:
        # ... and if everything else fails
        handle_except(e)
        exit_code = 1
    finally:
        # All cleanup actions are taken from here
        shutdown()
    return exit_code


# Now the sys.exit() calls are annoying: when main() calls
# sys.exit(), your interactive Python interpreter will exit!.
# The remedy is to let main()'s return value specify the
# exit status.
if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

"""
kaoru.commands.about
~~~~~~~~

/about command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import ParseMode
from .. import __version__ as pkg_version
from .. import PKG_URL as pkg_url
from .. import __name__ as pkg_name
from .. import __author__ as pkg_author
from .. import __copyright__ as pkg_copyright
from .. import config
from .. import utils
from . import bot_command

# /about command
@bot_command
def _cmd_handler(bot, update):
    """command handler"""

    # the about string
    about_str = \
    """
*{name} - v{version}*
{url}
    """.format(
    name=pkg_name,
    url=pkg_url,
    version=pkg_version,
    ).strip()

    # send the actual message
    utils.echo_msg(bot, update,
        about_str,
        parse_mode=ParseMode.MARKDOWN,
        disable_web_page_preview=True
    )

desc = 'Information for nerds'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'about'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.dryrun
~~~~~~~~

/dryrun command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from .. import config
from .. import utils
from . import bot_command

# /dryrun command
@bot_command
def _cmd_handler(bot, update):
    """toggle dry run mode"""
    config.set('dry_run', not config.get('dry_run'))
    if config.get('dry_run'):
        status = "ON"
    else:
        status = "OFF"
    utils.echo_msg(bot, update, "Dry run mode is {}".format(status))

desc = 'Toggle "dry run" mode'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'dryrun'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.unknown
~~~~~~~~

Default command handler

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from .. import utils
from . import bot_command

@bot_command
def _cmd_handler(bot, update):
    """default command handler"""

    chatter_name = update.message.from_user.first_name
    utils.echo_msg(
        bot, update,
        "I am sorry {}. I'm afraid I cannot do that ...".format(chatter_name)
    )

cmd_handler = _cmd_handler  # command handler



# -*- coding: utf-8 -*-

"""
kaoru.security
~~~~~~~~

Security routines

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from . import log
from . import config

class SecurityException(Exception):
    """Standard security exception"""
    pass

def check_masters(masters):
    if not len(masters):
        log.msg_warn("I have no master(s), I will only listen through console input (if set to true)")

def check_update(update):
    """Perform sanity checks on update"""

    # check whether the update comes from
    # one of my 'masters'
    if config.get('strict'):
        username = update.message.from_user.username
        command = update.message.text
        if not username in config.get('masters'):
            raise SecurityException("Unauthorized user '{}' attempted to run a command: {}".format(username, command))



# -*- coding: utf-8 -*-

"""
kaoru.commands.help
~~~~~~~~

/help command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from telegram import ParseMode
from .. import __version__ as pkg_version
from .. import PKG_URL as pkg_url
from .. import __name__ as pkg_name
from .. import __author__ as pkg_author
from .. import __copyright__ as pkg_copyright
from .. import config
from .. import utils
from . import bot_command


# /help command
@bot_command
def _cmd_handler(bot, update):
    # First of all, grab the current list of commands
    # and format them to Markdown
    commands = ""
    for command, desc, handler in cmd_list:
        commands += "/{} - {}\n".format(command, desc)

    # ... and construct the final markdown string for help
    help_str = \
    """
*The following is the list of available commands I have:*
---
{}
    """.format(commands).strip()

    # send the thing
    utils.echo_msg(bot, update, help_str, parse_mode=ParseMode.MARKDOWN)

cmd_list = None  #
desc = 'Get list of available commands'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'help'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.cli
~~~~~~~~

Command line interface

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from clint.textui import puts
from clint.textui.colored import red, cyan, yellow
from clint.textui.prompt import query

from . import command
from . import config

def _prompt_str(botname):
    """Print prompt string"""

    return str(cyan('{}>'.format(botname), bold=True))

def _cli_help(commands):
    puts()
    puts(yellow("Available bot commands are:", bold=True))
    puts(yellow("---------------------------", bold=True))
    for cmd, desc, handler in commands:
        puts(yellow("* /{:12} - {}".format(cmd, desc)))

    # additional commands go in here
    puts()
    puts(yellow("Additional commands", bold=True))
    puts(yellow("-------------------", bold=True))
    additional_cmds = [
       ('quit', 'exit this program'),
       ('help', 'print this help'),
    ]
    for cmd, desc in additional_cmds:
        puts(yellow("* {:12} - {}".format(cmd, desc)))
    puts()

def prompt_loop(dispatcher, update_queue):
    """Command line interface for the user"""

    puts(red("Entering CLI"))
    puts(red('Type "help" to get a list of available commands'))

    while True:
        # prompt the user for input
        cmd = query(_prompt_str(dispatcher.bot.username))

        # and then see what command came out of it
        if cmd == 'help':
            _cli_help(command.get_list())
            continue
        if cmd == 'quit':
            # Gracefully stop the event handler
            break

        # else, put the text into the update queue
        elif len(cmd ) > 0:
            update_queue.put(cmd )  # Put command into queue




# -*- coding: utf-8 -*-

"""
kaoru.commands.hibernate
~~~~~~~~

/hibernate command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import re

from .. import config
from .. import utils
from ..procutils import proc_exec_async, proc_select
from . import bot_command

# /hibernate command:
@bot_command
def _cmd_handler(bot, update):
    """hibernate your machine(s)"""

    hibernate_exec = proc_select(['pm-hibernate'])
    utils.echo_msg(
        bot, update,
        'Your host(s) are going to be put into HIBERNATION'
    )
    if re.match('.*pm-hibernate$', hibernate_exec):
        # pm-hibernate needs to be run as root
        hibernate_exec = 'sudo {}'.format(hibernate_exec)
        proc_exec_async(hibernate_exec)
    else:
        log.msg_err(
            "pm-hibernate has not been found, "
            "make sure you have installed pm-utils on your system."
        )


desc = 'Suspend your host(s) to disk'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'hibernate'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.screenshot
~~~~~~~~

/screenshot command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""
import os
import uuid
import re
from telegram import Update, ChatAction

from .. import utils
from .. import config
from .. import log
from . import bot_command
from ..procutils import proc_exec, proc_select

# a fun list of replies to the user
_replies = [
    'Here you go, Sir.',
    'Le voilà, Monsieur.',
    'As you wish.',
    'Sus deseos son órdenes para mi, mi Señor.'
]

# /screenshot command:
@bot_command
def _cmd_handler(bot, update):
    """
    it basically takes a screen shot
    """

    # check for executables set for commands
    screenshot_exec = proc_select(['import', 'scrot'])

    # file name is set without extension at first, depending on the
    # program being selected for the job, an extension will be chosen
    screenshot_file = "/tmp/{}".format(uuid.uuid4().hex)

    # screen shot quality (for jpeg)
    screenshot_jpeg_quality = 60

    # for the moment, there is only support
    # for scrot as the screenshooter

    # imagemagick
    if re.match('.*import$', screenshot_exec):
        screenshot_file += '.jpg'
        screenshot_exec = "{} -window root -quality {} {}".format(
            screenshot_exec,
            screenshot_jpeg_quality,
            screenshot_file
        )
    # scrot
    elif re.match('.*scrot$', screenshot_exec):
        screenshot_file += '.jpg'
        screenshot_exec = "{} -q {} {}".format(
            screenshot_exec,
            screenshot_jpeg_quality,
            screenshot_file
        )
    else:
        log.msg_err("No screenshot backend has been found!")

    # execute the thing
    if screenshot_exec is not None:
        if isinstance(update, Update):
            bot.sendChatAction(
                chat_id=update.message.chat_id,
                action=ChatAction.UPLOAD_PHOTO
            )
        proc_exec(screenshot_exec)

    # check is the file is available
    if os.path.isfile(screenshot_file):
        if isinstance(update, Update):
            with open(screenshot_file, 'rb') as photo:
                log.msg_debug("{}: {} byte(s)".format(
                    screenshot_file,
                    os.path.getsize(screenshot_file)
                ))
                log.msg_debug("{}: sending picture".format(screenshot_file))
                # send the actual pic
                bot.sendPhoto(
                    photo=photo,
                    chat_id=update.message.chat_id,
                    caption=utils.select_rand_str(_replies)
                )
                log.msg_debug("{}: picture sent".format(screenshot_file))
    else:
        utils.echo_msg(
            bot, update,
            "I am sorry, I have been unable "
            "to generate that screenshot you wished."
        )

desc = 'Get a screen shot from your host(s)'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'screenshot'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.cancel
~~~~~~~~

/cancel command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from .. import config
from .. import utils
from ..procutils import proc_exec_async
from . import bot_command

# /cancel command:
@bot_command
def _cmd_handler(bot, update):
    """Cancel any pending reboot/poweroffs"""

    proc_exec_async("sudo shutdown -c")
    config.set('queue_reboot', False)
    config.set('queue_poweroff', False)
    utils.echo_msg(bot, update, "Operations cancelled")

desc = 'Cancel any pending operation(s)'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'cancel'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.commands.reboot
~~~~~~~~

/reboot command implementation

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

from .. import config
from .. import utils
from ..procutils import proc_exec_async
from . import bot_command

# /reboot command:
@bot_command
def _cmd_handler(bot, update):
    """reboot your machine(s)"""

    if not config.get('queue_poweroff') and not config.get('queue_reboot'):
        utils.echo_msg(
            bot, update,
            "I'm going to be rebooted in about {} minute(s)".format(
                config.get('reboot_delay')
            )
        )
        proc_exec_async("sudo shutdown -r +{}".format(
                config.get('reboot_delay')
            )
        )
        config.set('queue_reboot', True)
    else:
        utils.echo_msg(
            bot, update,
            "Your machines are already going to be shutdown/rebooted"
        )

desc = 'Reboot your host(s)'  # This command's description
cmd_handler = _cmd_handler  # command handler
cmd_str = 'reboot'  # command /string



# -*- coding: utf-8 -*-

"""
kaoru.db
~~~~~~~~

sqlite3 database module

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""

import sqlite3
import os
import time

from . import log

#################
# Database schema
#################
_db_schema = """
CREATE TABLE updates (
   id      INTEGER PRIMARY KEY NOT NULL,
   usr_id  INTEGER NOT NULL,
   message TEXT NOT NULL
)
"""
_db_file = None # database file used by sqlite3

def _create_schema(db):
    """Generate new sqlite schema on database file"""
    log.msg_debug("Creating new sqlite3 db at {}".format(db))
    with sqlite3.connect(db) as conn:
        cursor = conn.cursor()
        cursor.execute(_db_schema)

def init(*, db):
    """Initialise database interface

    :param db: database file to use
    """

    # check whether the database file actually exists
    if not os.path.isfile(db):
        log.msg_warn("Database not found!, creating a new one ...")
        # create schema
        _create_schema(db)

    global _db_file
    _db_file = db

def insert_update(update):
    """Insert Telegram update into database"""

    update_id = update.update_id
    user_id = update.message.from_user.id
    message = update.message.text
    log.msg_debug("sqlite: inserting update '{}'".format(update_id))
    query("INSERT INTO updates VALUES({}, {}, '{}')".format(
            update_id, user_id, message
        )
    )

def get_last_update_id():
    """Get latest update_id"""

    rows = query("SELECT * FROM updates ORDER BY id DESC LIMIT 1")
    if len(rows):
        return rows[0][0]
    return None

def query(sql):
    """Perform query on database"""

    log.msg_debug("sqlite: connecting to database")
    rows = None
    start_time = time.time()
    with sqlite3.connect(_db_file) as conn:
        cursor = conn.cursor()
        log.msg_debug("sqlite: {}".format(sql))
        cursor.execute(sql)
        dt = time.time() - start_time
        log.msg_debug("sqlite: done ({:.3f} ms)".format(dt*1000))
        rows = cursor.fetchall()
        if len(rows):
            log.msg_debug("sqlite: got {} row(s)".format(len(rows)))
        else:
            log.msg_debug("sqlite: got nothing from this query")
    return rows




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
setuptools config file
"""

import sys
import pip
from pip.req import parse_requirements
from setuptools import setup, find_packages
import os
from kaoru import __version__ as version
from kaoru import __author__ as author
from kaoru import PKG_URL as pkg_url
from kaoru import __name__ as pkg_name


# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session=pip.download.PipSession())

# reqs is a list of requirement
# e.g. ['django==1.5.1', 'mezzanine==1.4.6']
reqs = [str(ir.req) for ir in install_reqs]
desc = "A Telegram Bot as your personal IoT assistant"

setup(
    name=pkg_name,
    version=version,
    packages=find_packages(),
    author=author,
    author_email="alejandroricoveri@gmail.com",
    description=desc,
    long_description=open("README.rst").read(),
    url=pkg_url,
    license='MIT',
    download_url="{url}/tarball/{version}".format(url=pkg_url, version=version),
    keywords=['telegram', 'iot'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console	',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.3',
    ],
    entry_points={
        'console_scripts': [
            'kaoru = kaoru.__main__:main',
        ],
    },
    install_requires = reqs,
)



# -*- coding: utf-8 -*-

"""
kaoru.log
~~~~~~~~~~~~~

Handles local and system-wide logging into files.
System-wide messages are logged through logging.

:copyright: (c) 2015 by Alejandro Ricoveri
:license: MIT, see LICENSE for more details.

"""


import sys
import logging
import logging.handlers
from . import __name__ as pkg_name
from clint.textui.colored import white, red, cyan, yellow, green
from clint.textui import puts

#
# Constants
#
LOG_LVL_DEFAULT = 0
LOG_FILE_DEFAULT = "{pkg_name}.log".format(pkg_name=pkg_name)

# Globals
_logger = None
_log_lvl = 0
_stdout = False


def _set_lvl(lvl):
    if lvl == 0:
        return logging.DEBUG
    elif lvl == 1:
        return logging.INFO
    elif lvl == 2:
        return logging.WARNING
    elif lvl == 3:
        return logging.ERROR
    elif lvl >= 4:
        return logging.CRITICAL
    else:
        return logging.INFO


def init(*, threshold_lvl=1, quiet_stdout=False, log_file):
    """
    Initiate the log module

    :param threshold_lvl: messages under this level won't be issued/logged
    :param to_stdout: activate stdout log stream
    """
    global _logger, _log_lvl

    # translate lvl to those used by 'logging' module
    _log_lvl = _set_lvl(threshold_lvl)

    # logger Creation
    _logger = logging.getLogger(pkg_name)
    _logger.setLevel(_log_lvl)

    # create syslog handler and set level to info
    log_h = logging.FileHandler(log_file)

    # Base message format
    base_fmt = '%(asctime)s - %(name)s - [%(levelname)s] - %(message)s'

    # set formatter
    log_fmt = logging.Formatter(base_fmt)
    log_h.setFormatter(log_fmt)
    # add Handler
    _logger.addHandler(log_h)

    # create stout handler
    if not quiet_stdout:
        global _stdout
        _stdout = True


def get_logger():
    """get logger interface for external use"""
    return _logger


def to_stdout(msg, *, colorf=green, bold=False):
    if _stdout:
        puts(colorf(msg, bold=bold))


def msg(message, *, bold=False):
    """
    Log a regular message

    :param message: the message to be logged
    """
    to_stdout(" --- {message}".format(message=message), bold=bold)
    if _logger:
        _logger.info(message)


def msg_warn(message):
    """
    Log a warning message

    :param message: the message to be logged
    """
    to_stdout(" (!) {message}".format(message=message),
              colorf=yellow, bold=True)
    if _logger:
        _logger.warn(message)


def msg_err(message):
    """
    Log an error message

    :param message: the message to be logged
    """
    to_stdout(" !!! {message}".format(message=message), colorf=red, bold=True)
    if _logger:
        _logger.error(message)


def msg_debug(message):
    """
    Log a debug message

    :param message: the message to be logged
    """
    if _log_lvl == logging.DEBUG:
        to_stdout(" (*) {message}".format(message=message), colorf=cyan)
        if _logger:
            _logger.debug(message)
